import logging
import unittest

from destral.openerp import OpenERPService
from destral.transaction import Transaction
from osconf import config_from_environment


logger = logging.getLogger('destral.testing')


class OOTestCase(unittest.TestCase):
    """Base class to inherit test cases from for OpenERP Testing Framework.
    """

    @property
    def database(self):
        return self.openerp.db_name

    def setUp(self):
        self.config = config_from_environment('DESTRAL', ['module'])
        self.openerp = OpenERPService()
        self.drop_database = False
        if not self.openerp.db_name:
            self.openerp.db_name = self.openerp.create_database()
            self.drop_database = True
            self.openerp.install_module(self.config['module'])

    def test_all_views(self):
        logger.info('Testing views for module %s', self.config['module'])
        imd_obj = self.openerp.pool.get('ir.model.data')
        view_obj = self.openerp.pool.get('ir.ui.view')
        with Transaction().start(self.database) as txn:
            imd_ids = imd_obj.search(txn.cursor, txn.user, [
                ('model', '=', 'ir.ui.view'),
                ('module', '=', self.config['module'])
            ])
            if imd_ids:
                views = {}
                for imd in imd_obj.read(txn.cursor, txn.user, imd_ids):
                    view_xml_name = '{}.{}'.format(imd['module'], imd['name'])
                    views[imd['res_id']] = view_xml_name
                view_ids = views.keys()
                logger.info('Testing %s views...', len(view_ids))
                for view in view_obj.browse(txn.cursor, txn.user, view_ids):
                    view_xml_name = views[view.id]
                    model = self.openerp.pool.get(view.model)
                    if model is None:
                        # Check if model exists
                        raise Exception(
                            'View (xml id: %s) references model %s which does '
                            'not exist' % (view_xml_name, view.model)
                        )
                    logger.info('Testing view %s (id: %s)', view.name, view.id)
                    model.fields_view_get(txn.cursor, txn.user, view.id,
                                          view.type)

    def tearDown(self):
        if self.drop_database:
            self.openerp.drop_database()
        self.openerp.config['db_name'] = False



import logging
import time

from osconf import config_from_environment
from destral.utils import update_config


logger = logging.getLogger('destral.openerp')
DEFAULT_USER = 1


def patched_pool_jobs(*args, **kwargs):
    logger.debug('Patched ir.cron')
    return False


class OpenERPService(object):

    def __init__(self, **kwargs):
        config = config_from_environment('OPENERP', [], **kwargs)
        import netsvc
        import tools
        update_config(tools.config, **config)
        tools.config.parse()
        from tools import config as default_config
        self.config = update_config(default_config, **config)
        import pooler
        import workflow
        self.pooler = pooler
        self.db = None
        self.pool = None
        if 'db_name' in config:
            self.db_name = config['db_name']

    def create_database(self):
        db_name = 'test_' + str(int(time.time()))
        import sql_db
        conn = sql_db.db_connect('template1')
        cursor = conn.cursor()
        try:
            logger.info('Creating database %s', db_name)
            cursor.autocommit(True)
            cursor.execute('CREATE DATABASE ' + db_name + ' WITH TEMPLATE base')
            return db_name
        finally:
            cursor.close()

    def drop_database(self):
        import sql_db
        sql_db.close_db(self.db_name)
        conn = sql_db.db_connect('template1')
        cursor = conn.cursor()
        try:
            logger.info('Droping database %s', self.db_name)
            cursor.autocommit(True)
            cursor.execute('DROP DATABASE ' + self.db_name)
        finally:
            cursor.close()

    @property
    def db_name(self):
        return self.config['db_name']

    @db_name.setter
    def db_name(self, value):
        self.config['db_name'] = value
        self.db, self.pool = self.pooler.get_db_and_pool(self.db_name)
        logger.debug('Patching ir.cron _poolJobs with %s', patched_pool_jobs)
        cron = self.pool.get('ir.cron')
        cron._poolJobs = patched_pool_jobs
        self.pool.obj_pool['ir.cron'] = cron

    def install_module(self, module):
        logger.info('Installing module %s', module)
        import pooler
        from destral.transaction import Transaction
        module_obj = self.pool.get('ir.module.module')
        with Transaction().start(self.config['db_name']) as txn:
            module_obj.update_list(txn.cursor, txn.user)
            module_ids = module_obj.search(
                txn.cursor, DEFAULT_USER,
                [('name', '=', module)],
            )
            assert module_ids, "Module %s not found" % module
            module_obj.button_install(txn.cursor, DEFAULT_USER, module_ids)
            pool = pooler.get_pool(txn.cursor.dbname)
            mod_obj = pool.get('ir.module.module')
            ids = mod_obj.search(txn.cursor, txn.user, [
                ('state', 'in', ['to upgrade', 'to remove', 'to install'])
            ])
            unmet_packages = []
            mod_dep_obj = pool.get('ir.module.module.dependency')
            for mod in mod_obj.browse(txn.cursor, txn.user, ids):
                deps = mod_dep_obj.search(txn.cursor, txn.user, [
                    ('module_id', '=', mod.id)
                ])
                for dep_mod in mod_dep_obj.browse(txn.cursor, txn.user, deps):
                    if dep_mod.state in ('unknown', 'uninstalled'):
                        unmet_packages.append(dep_mod.name)
            mod_obj.download(txn.cursor, txn.user, ids)
            txn.cursor.commit()
        self.db, self.pool = pooler.restart_pool(
            self.config['db_name'], update_module=True
        )



import os


def update_config(config, **kwargs):
    for key, value in kwargs.iteritems():
        config[key] = value
    return config


def detect_module(path):
    stack = path.split(os.path.sep)
    if not stack[0]:
        stack[0] = os.path.sep
    stack = [x for x in stack if x]
    while stack:
        path = os.path.join(*stack)
        module = stack.pop()
        if not os.path.isdir(path):
            continue
        files = os.listdir(path)
        if '__terp__.py' in files:
            return module
    return None


from threading import local


from destral.openerp import OpenERPService


class Singleton(type):
    '''
    Metaclass for singleton pattern
    :copyright: Tryton Project
    '''
    def __init__(mcs, name, bases, dict_):
        super(Singleton, mcs).__init__(name, bases, dict_)
        mcs.instance = None

    def __call__(mcs, *args, **kwargs):
        if mcs.instance is None:
            mcs.instance = super(Singleton, mcs).__call__(*args, **kwargs)
        return mcs.instance


class Transaction(local):
    __metaclass__ = Singleton

    database = None
    service = None
    pool = None
    cursor = None
    user = None
    context = None

    def __init__(self):
        pass

    def start(self, database_name, user=1, context=None):
        self._assert_stopped()
        self.service = OpenERPService(db_name=database_name)
        self.pool = self.service.pool
        self.cursor = self.service.db.cursor()
        self.user = user
        self.context = context if context is not None else self.get_context()
        return self

    def stop(self):
        'End the transaction'
        self.cursor.close()
        self.service = None
        self.cursor = None
        self.user = None
        self.context = None
        self.database = None
        self.pool = None

    def get_context(self):
        'Loads the context of the current user'
        assert self.user is not None

        user_obj = self.pool.get('res.users')
        return user_obj.context_get(self.cursor, self.user)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.stop()

    def _assert_stopped(self):
        'Assert that there is no active transaction'
        assert self.service is None
        assert self.database is None
        assert self.cursor is None
        assert self.pool is None
        assert self.user is None
        assert self.context is None